package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.Iterator;
import java.util.HashSet;
import java.util.Set;
import java.util.HashMap;

class BoundSet
{
    static final BoundSet TRUE;
    static final BoundSet FALSE;
    HashMap<InferenceVariable, BoundSet.ThreeSets> boundsPerVariable;
    HashMap<ParameterizedTypeBinding, ParameterizedTypeBinding> captures;
    Set<InferenceVariable> inThrows;
    private TypeBound[] incorporatedBounds;
    private TypeBound[] unincorporatedBounds;
    private int unincorporatedBoundsCount;
    private TypeBound[] mostRecentBounds;
    
    static {
        TRUE = new BoundSet();
        FALSE = new BoundSet();
    }
    
    public BoundSet() {
        this.boundsPerVariable = new HashMap();
        this.captures = new HashMap();
        this.inThrows = new HashSet();
        this.incorporatedBounds = new TypeBound[0];
        this.unincorporatedBounds = new TypeBound[1024];
        this.unincorporatedBoundsCount = 0;
        this.mostRecentBounds = new TypeBound[4];
    }
    
    public void addBoundsFromTypeParameters(InferenceContext18 context, TypeVariableBinding[] typeParameters, InferenceVariable[] variables) {
        for (int length = typeParameters.length, i = 0; i < length; ++i) {
            TypeVariableBinding typeParameter = typeParameters[i];
            InferenceVariable variable = variables[i];
            TypeBound[] someBounds = typeParameter.getTypeBounds(variable, new InferenceSubstitution(context));
            boolean hasProperBound = false;
            if (someBounds.length > 0) {
                hasProperBound = this.addBounds(someBounds, context.environment);
            }
            if (!hasProperBound) {
                this.addBound(new TypeBound(variable, context.object, 2), context.environment);
            }
        }
    }
    
    public TypeBound[] flatten() {
        int size = 0;
        Iterator<BoundSet.ThreeSets> outerIt = this.boundsPerVariable.values().iterator();
        while (outerIt.hasNext()) {
            size += ((BoundSet.ThreeSets)outerIt.next()).size();
        }
        TypeBound[] collected = new TypeBound[size];
        if (size == 0) {
            return collected;
        }
        outerIt = this.boundsPerVariable.values().iterator();
        int idx = 0;
        while (outerIt.hasNext()) {
            idx = ((BoundSet.ThreeSets)outerIt.next()).flattenInto(collected, idx);
        }
        return collected;
    }
    
    public BoundSet copy() {
        BoundSet copy = new BoundSet();
        for (Map.Entry<InferenceVariable, BoundSet.ThreeSets> entry : this.boundsPerVariable.entrySet()) {
            copy.boundsPerVariable.put((InferenceVariable)entry.getKey(), ((BoundSet.ThreeSets)entry.getValue()).copy());
        }
        copy.inThrows.addAll(this.inThrows);
        copy.captures.putAll(this.captures);
        System.arraycopy(this.incorporatedBounds, 0, copy.incorporatedBounds = new TypeBound[this.incorporatedBounds.length], 0, this.incorporatedBounds.length);
        System.arraycopy(this.unincorporatedBounds, 0, copy.unincorporatedBounds = new TypeBound[this.unincorporatedBounds.length], 0, this.unincorporatedBounds.length);
        copy.unincorporatedBoundsCount = this.unincorporatedBoundsCount;
        return copy;
    }
    
    public void addBound(TypeBound bound, LookupEnvironment environment) {
        if (bound.relation == 2 && bound.right.id == 1) {
            return;
        }
        if (bound.left == bound.right) {
            return;
        }
        int recent = 0;
        while (recent < 4) {
            if (bound.equals(this.mostRecentBounds[recent])) {
                if (environment.globalOptions.isAnnotationBasedNullAnalysisEnabled) {
                    TypeBound existing = this.mostRecentBounds[recent];
                    long boundNullBits = bound.right.tagBits & 0x180000000000000L;
                    long existingNullBits = existing.right.tagBits & 0x180000000000000L;
                    if (boundNullBits != existingNullBits) {
                        if (existingNullBits == 0L) {
                            existing.right = bound.right;
                        }
                        else if (boundNullBits != 0L) {
                            existing.right = environment.createAnnotatedType(existing.right, environment.nullAnnotationsFromTagBits(boundNullBits));
                        }
                    }
                }
                return;
            }
            else {
                ++recent;
            }
        }
        this.mostRecentBounds[3] = this.mostRecentBounds[2];
        this.mostRecentBounds[2] = this.mostRecentBounds[1];
        this.mostRecentBounds[1] = this.mostRecentBounds[0];
        this.mostRecentBounds[0] = bound;
        InferenceVariable variable = bound.left.prototype();
        BoundSet.ThreeSets three = (BoundSet.ThreeSets)this.boundsPerVariable.get(variable);
        if (three == null) {
            this.boundsPerVariable.put(variable, three = new BoundSet.ThreeSets());
        }
        if (three.addBound(bound)) {
            int unincorporatedBoundsLength = this.unincorporatedBounds.length;
            if (this.unincorporatedBoundsCount >= unincorporatedBoundsLength) {
                System.arraycopy(this.unincorporatedBounds, 0, this.unincorporatedBounds = new TypeBound[unincorporatedBoundsLength * 2], 0, unincorporatedBoundsLength);
            }
            this.unincorporatedBounds[this.unincorporatedBoundsCount++] = bound;
            TypeBinding typeBinding = bound.right;
            if (bound.relation == 4 && typeBinding.isProperType(true)) {
                three.setInstantiation(typeBinding, variable, environment);
            }
            if ((bound.right instanceof InferenceVariable)) {
                InferenceVariable rightIV = (InferenceVariable)bound.right.prototype();
                three = (BoundSet.ThreeSets)this.boundsPerVariable.get(rightIV);
                if (three == null) {
                    this.boundsPerVariable.put(rightIV, three = new BoundSet.ThreeSets());
                }
                if (three.inverseBounds == null) {
                    three.inverseBounds = new HashMap();
                }
                three.inverseBounds.put(rightIV, bound);
            }
        }
        return;
    }
    
    private boolean addBounds(TypeBound[] newBounds, LookupEnvironment environment) {
        boolean hasProperBound = false;
        for (int i = 0; i < newBounds.length; hasProperBound |= newBounds[i].isBound(), ++i) {
            this.addBound(newBounds[i], environment);
        }
        return hasProperBound;
    }
    
    public void addBounds(BoundSet that, LookupEnvironment environment) {
        if (that == null || environment == null) {
            return;
        }
        this.addBounds(that.flatten(), environment);
        return;
    }
    
    public boolean isInstantiated(InferenceVariable inferenceVariable) {
        BoundSet.ThreeSets three = (BoundSet.ThreeSets)this.boundsPerVariable.get(inferenceVariable.prototype());
        return three != null && three.instantiation != null;
    }
    
    public TypeBinding getInstantiation(InferenceVariable inferenceVariable, LookupEnvironment environment) {
        BoundSet.ThreeSets three = (BoundSet.ThreeSets)this.boundsPerVariable.get(inferenceVariable.prototype());
        if (three != null) {
            TypeBinding instantiation = three.instantiation;
            if (environment != null && environment.globalOptions.isAnnotationBasedNullAnalysisEnabled && instantiation != null && (instantiation.tagBits & 0x180000000000000L) == 0x0L) {
                return three.combineAndUseNullHints(instantiation, inferenceVariable.nullHints, environment);
            }
            return instantiation;
        }
        else {
            return null;
        }
    }
    
    public int numUninstantiatedVariables(InferenceVariable[] variables) {
        int num = 0;
        for (int i = 0; i < variables.length; ++i) {
            if (!this.isInstantiated(variables[i])) {
                ++num;
            }
        }
        return num;
    }
    
    boolean incorporate(InferenceContext18 context) throws InferenceFailureException {
        if (this.unincorporatedBoundsCount == 0 && this.captures.size() == 0) {
            return true;
        }
        do {
            TypeBound[] freshBounds;
            System.arraycopy(this.unincorporatedBounds, 0, freshBounds = new TypeBound[this.unincorporatedBoundsCount], 0, this.unincorporatedBoundsCount);
            this.unincorporatedBoundsCount = 0;
            if (!this.incorporate(context, this.incorporatedBounds, freshBounds)) {
                return false;
            }
            if (!this.incorporate(context, freshBounds, freshBounds)) {
                return false;
            }
            int incorporatedLength = this.incorporatedBounds.length;
            int unincorporatedLength = freshBounds.length;
            TypeBound[] aggregate = new TypeBound[incorporatedLength + unincorporatedLength];
            System.arraycopy(this.incorporatedBounds, 0, aggregate, 0, incorporatedLength);
            System.arraycopy(freshBounds, 0, aggregate, incorporatedLength, unincorporatedLength);
            this.incorporatedBounds = aggregate;
        } while (this.unincorporatedBoundsCount > 0);
        return true;
    }
    
    boolean incorporate(InferenceContext18 context, TypeBound[] first, TypeBound[] next) throws InferenceFailureException {
        boolean analyzeNull = context.environment.globalOptions.isAnnotationBasedNullAnalysisEnabled;
        ConstraintTypeFormula[] mostRecentFormulas = new ConstraintTypeFormula[4];
        for (int i = 0, iLength = first.length; i < iLength; ++i) {
            TypeBound boundI = first[i];
            for (int j = 0, jLength = next.length; j < jLength; ++j) {
                TypeBound boundJ = next[j];
                if (boundI != boundJ) {
                    int iteration = 1;
                    do {
                        ConstraintTypeFormula newConstraint = null;
                        boolean deriveTypeArgumentConstraints = false;
                        if (iteration == 2) {
                            TypeBound boundX = boundI;
                            boundI = boundJ;
                            boundJ = boundX;
                        }
                        switch (boundI.relation) {
                            case 4: {
                                switch (boundJ.relation) {
                                    case 4: {
                                        newConstraint = this.combineSameSame(boundI, boundJ);
                                        break;
                                    }
                                    case 2:
                                    case 3: {
                                        newConstraint = this.combineSameSubSuper(boundI, boundJ);
                                        break;
                                    }
                                }
                                break;
                            }
                            case 2: {
                                switch (boundJ.relation) {
                                    case 4: {
                                        newConstraint = this.combineSameSubSuper(boundJ, boundI);
                                        break;
                                    }
                                    case 3: {
                                        newConstraint = this.combineSuperAndSub(boundJ, boundI);
                                        break;
                                    }
                                    case 2: {
                                        newConstraint = this.combineEqualSupers(boundI, boundJ);
                                        deriveTypeArgumentConstraints = TypeBinding.equalsEquals(boundI.left, boundJ.left);
                                        break;
                                    }
                                }
                                break;
                            }
                            case 3: {
                                switch (boundJ.relation) {
                                    case 4: {
                                        newConstraint = this.combineSameSubSuper(boundJ, boundI);
                                        break;
                                    }
                                    case 2: {
                                        newConstraint = this.combineSuperAndSub(boundI, boundJ);
                                        break;
                                    }
                                    case 3: {
                                        newConstraint = this.combineEqualSupers(boundI, boundJ);
                                        break;
                                    }
                                }
                                break;
                            }
                        }
                        if (newConstraint != null) {
                            if (newConstraint.left == newConstraint.right) {
                                newConstraint = null;
                            }
                            else if (newConstraint.equalsEquals(mostRecentFormulas[0]) || newConstraint.equalsEquals(mostRecentFormulas[1]) || newConstraint.equalsEquals(mostRecentFormulas[2]) || newConstraint.equalsEquals(mostRecentFormulas[3])) {
                                newConstraint = null;
                            }
                        }
                        if (newConstraint != null) {
                            mostRecentFormulas[3] = mostRecentFormulas[2];
                            mostRecentFormulas[2] = mostRecentFormulas[1];
                            mostRecentFormulas[1] = mostRecentFormulas[0];
                            mostRecentFormulas[0] = newConstraint;
                            if (!this.reduceOneConstraint(context, newConstraint)) {
                                return false;
                            }
                            if (analyzeNull) {
                                long nullHints = (newConstraint.left.tagBits | newConstraint.right.tagBits) & 0x180000000000000L;
                                if (nullHints != 0L && (TypeBinding.equalsEquals(boundI.left, boundJ.left) || (boundI.relation == 4 && TypeBinding.equalsEquals(boundI.right, boundJ.left)) || (boundJ.relation == 4 && TypeBinding.equalsEquals(boundI.left, boundJ.right)))) {
                                    boundI.nullHints |= nullHints;
                                    boundJ.nullHints |= nullHints;
                                }
                            }
                        }
                        ConstraintFormula[] typeArgumentConstraints = deriveTypeArgumentConstraints ? this.deriveTypeArgumentConstraints(boundI, boundJ) : null;
                        if (typeArgumentConstraints != null) {
                            int k = 0;
                            int length = typeArgumentConstraints.length;
                            while (k < length) {
                                if (!this.reduceOneConstraint(context, typeArgumentConstraints[k])) {
                                    return false;
                                }
                                ++k;
                            }
                        }
                        if (iteration == 2) {
                            TypeBound boundX2 = boundI;
                            boundI = boundJ;
                            boundJ = boundX2;
                        }
                        else {
                            continue;
                        }
                    } while (first != next && ++iteration <= 2);
                }
            }
        }
        for (Map.Entry<ParameterizedTypeBinding, ParameterizedTypeBinding> capt : this.captures.entrySet()) {
            ParameterizedTypeBinding gAlpha = (ParameterizedTypeBinding)capt.getKey();
            ParameterizedTypeBinding gA = (ParameterizedTypeBinding)capt.getValue();
            ReferenceBinding g = (ReferenceBinding)gA.original();
            final TypeVariableBinding[] parameters = g.typeVariables();
            final InferenceVariable[] alphas = new InferenceVariable[gAlpha.arguments.length];
            System.arraycopy(gAlpha.arguments, 0, alphas, 0, alphas.length);
            InferenceSubstitution theta = new InferenceSubstitution(this, context.environment, alphas, context.currentInvocation, parameters) {
                
                @Override
                protected TypeBinding getP(int i) {
                    return parameters[i];
                }
            };
            for (int l = 0, length2 = parameters.length; l < length2; ++l) {
                TypeVariableBinding pi = parameters[l];
                InferenceVariable alpha = (InferenceVariable)gAlpha.arguments[l];
                this.addBounds(pi.getTypeBounds(alpha, theta), context.environment);
                TypeBinding ai = gA.arguments[l];
                if ((ai instanceof WildcardBinding)) {
                    WildcardBinding wildcardBinding = (WildcardBinding)ai;
                    TypeBinding t = wildcardBinding.bound;
                    BoundSet.ThreeSets three = (BoundSet.ThreeSets)this.boundsPerVariable.get(alpha.prototype());
                    if (three != null) {
                        if (three.sameBounds != null) {
                            for (TypeBound bound : three.sameBounds) {
                                if (bound.right instanceof CaptureBinding && bound.right.isProperType(true)) {
                                    continue;
                                }
                                if (!(bound.right instanceof InferenceVariable)) {
                                    return false;
                                }
                                continue;
                            }
                        }
                        if (three.subBounds != null && pi.firstBound != null) {
                            for (TypeBound bound : three.subBounds) {
                                if (!(bound.right instanceof InferenceVariable)) {
                                    TypeBinding r = bound.right;
                                    TypeBinding bi1 = pi.firstBound;
                                    ReferenceBinding[] otherBounds = pi.superInterfaces;
                                    TypeBinding bi2;
                                    if (otherBounds == Binding.NO_SUPERINTERFACES) {
                                        bi2 = bi1;
                                    }
                                    else {
                                        int n = otherBounds.length + 1;
                                        ReferenceBinding[] allBounds = new ReferenceBinding[n];
                                        allBounds[0] = (ReferenceBinding)bi1;
                                        System.arraycopy(otherBounds, 0, allBounds, 1, n - 1);
                                        bi2 = context.environment.createIntersectionType18(allBounds);
                                    }
                                    this.addTypeBoundsFromWildcardBound(context, theta, wildcardBinding.boundKind, t, r, bi2);
                                }
                                else {
                                    continue;
                                }
                            }
                        }
                        if (three.superBounds != null) {
                            for (TypeBound bound : three.superBounds) {
                                if (!(bound.right instanceof InferenceVariable)) {
                                    if (wildcardBinding.boundKind == 2) {
                                        this.reduceOneConstraint(context, ConstraintTypeFormula.create(bound.right, t, 2));
                                    }
                                    else {
                                        return false;
                                    }
                                }
                                else {
                                    continue;
                                }
                            }
                        }
                    }
                }
                else {
                    this.addBound(new TypeBound(alpha, ai, 4), context.environment);
                }
            }
        }
        this.captures.clear();
        return true;
    }
    
    void addTypeBoundsFromWildcardBound(InferenceContext18 context, InferenceSubstitution theta, int boundKind, TypeBinding t, TypeBinding r, TypeBinding bi) throws InferenceFailureException {
        ConstraintFormula formula = null;
        if (boundKind == 1) {
            if (bi.id == 1) {
                formula = ConstraintTypeFormula.create(t, r, 2);
            }
            if (t.id == 1) {
                formula = ConstraintTypeFormula.create(theta.substitute(theta, bi), r, 2);
            }
        }
        else {
            formula = ConstraintTypeFormula.create(theta.substitute(theta, bi), r, 2);
        }
        if (formula != null) {
            this.reduceOneConstraint(context, formula);
        }
    }
    
    private ConstraintTypeFormula combineSameSame(TypeBound boundS, TypeBound boundT) {
        if (TypeBinding.equalsEquals(boundS.left, boundT.left)) {
            return ConstraintTypeFormula.create(boundS.right, boundT.right, 4, boundS.isSoft || boundT.isSoft);
        }
        ConstraintTypeFormula newConstraint = this.combineSameSameWithProperType(boundS, boundT);
        if (newConstraint != null) {
            return newConstraint;
        }
        newConstraint = this.combineSameSameWithProperType(boundT, boundS);
        if (newConstraint != null) {
            return newConstraint;
        }
        return null;
    }
    
    private ConstraintTypeFormula combineSameSameWithProperType(TypeBound boundLeft, TypeBound boundRight) {
        TypeBinding u = boundLeft.right;
        if (u.isProperType(true)) {
            InferenceVariable alpha = boundLeft.left;
            TypeBinding left = boundRight.left;
            TypeBinding right = boundRight.right.substituteInferenceVariable(alpha, u);
            return ConstraintTypeFormula.create(left, right, 4, boundLeft.isSoft || boundRight.isSoft);
        }
        return null;
    }
    
    private ConstraintTypeFormula combineSameSubSuper(TypeBound boundS, TypeBound boundT) {
        InferenceVariable alpha = boundS.left;
        TypeBinding s = boundS.right;
        if (TypeBinding.equalsEquals(alpha, boundT.left)) {
            TypeBinding t = boundT.right;
            return ConstraintTypeFormula.create(s, t, boundT.relation, boundT.isSoft || boundS.isSoft);
        }
        if (TypeBinding.equalsEquals(alpha, boundT.right)) {
            TypeBinding t = boundT.left;
            return ConstraintTypeFormula.create(t, s, boundT.relation, boundT.isSoft || boundS.isSoft);
        }
        if ((boundS.right instanceof InferenceVariable)) {
            alpha = (InferenceVariable)boundS.right;
            s = boundS.left;
            if (TypeBinding.equalsEquals(alpha, boundT.left)) {
                TypeBinding t = boundT.right;
                return ConstraintTypeFormula.create(s, t, boundT.relation, boundT.isSoft || boundS.isSoft);
            }
            if (TypeBinding.equalsEquals(alpha, boundT.right)) {
                TypeBinding t = boundT.left;
                return ConstraintTypeFormula.create(t, s, boundT.relation, boundT.isSoft || boundS.isSoft);
            }
        }
        TypeBinding u = boundS.right;
        if (u.isProperType(true)) {
            boolean substitute = TypeBinding.equalsEquals(alpha, boundT.left);
            TypeBinding left = substitute ? u : boundT.left;
            TypeBinding right = boundT.right.substituteInferenceVariable(alpha, u);
            substitute |= TypeBinding.notEquals(right, boundT.right);
            if (substitute) {
                return ConstraintTypeFormula.create(left, right, boundT.relation, boundT.isSoft || boundS.isSoft);
            }
        }
        return null;
    }
    
    private ConstraintTypeFormula combineSuperAndSub(TypeBound boundS, TypeBound boundT) {
        InferenceVariable alpha = boundS.left;
        if (TypeBinding.equalsEquals(alpha, boundT.left)) {
            return ConstraintTypeFormula.create(boundS.right, boundT.right, 2, boundT.isSoft || boundS.isSoft);
        }
        if ((boundS.right instanceof InferenceVariable)) {
            alpha = (InferenceVariable)boundS.right;
            if (TypeBinding.equalsEquals(alpha, boundT.right)) {
                return ConstraintTypeFormula.create(boundS.left, boundT.left, 3, boundT.isSoft || boundS.isSoft);
            }
        }
        return null;
    }
    
    private ConstraintTypeFormula combineEqualSupers(TypeBound boundS, TypeBound boundT) {
        if (TypeBinding.equalsEquals(boundS.left, boundT.right)) {
            return ConstraintTypeFormula.create(boundT.left, boundS.right, boundS.relation, boundT.isSoft || boundS.isSoft);
        }
        if (TypeBinding.equalsEquals(boundS.right, boundT.left)) {
            return ConstraintTypeFormula.create(boundS.left, boundT.right, boundS.relation, boundT.isSoft || boundS.isSoft);
        }
        return null;
    }
    
    private ConstraintTypeFormula[] deriveTypeArgumentConstraints(TypeBound boundS, TypeBound boundT) {
        TypeBinding[] supers = this.superTypesWithCommonGenericType(boundS.right, boundT.right);
        if (supers != null) {
            return this.typeArgumentEqualityConstraints(supers[0], supers[1], boundS.isSoft || boundT.isSoft);
        }
        return null;
    }
    
    private ConstraintTypeFormula[] typeArgumentEqualityConstraints(TypeBinding s, TypeBinding t, boolean isSoft) {
        if (s == null || s.kind() != 260 || t == null || t.kind() != 260) {
            return null;
        }
        if (TypeBinding.equalsEquals(s, t)) {
            return null;
        }
        TypeBinding[] sis = s.typeArguments();
        TypeBinding[] tis = t.typeArguments();
        if (sis == null || tis == null || sis.length != tis.length) {
            return null;
        }
        List<ConstraintTypeFormula> result = new ArrayList();
        for (int i = 0; i < sis.length; ++i) {
            TypeBinding si = sis[i];
            TypeBinding ti = tis[i];
            if (!si.isWildcard() && !ti.isWildcard() && !TypeBinding.equalsEquals(si, ti)) {
                result.add(ConstraintTypeFormula.create(si, ti, 4, isSoft));
            }
        }
        if (result.size() > 0) {
            return (ConstraintTypeFormula[])result.toArray(new ConstraintTypeFormula[result.size()]);
        }
        return null;
    }
    
    public boolean reduceOneConstraint(InferenceContext18 context, ConstraintFormula currentConstraint) throws InferenceFailureException {
        Object result = currentConstraint.reduce(context);
        if (result == ReductionResult.FALSE) {
            return false;
        }
        if (result == ReductionResult.TRUE) {
            return true;
        }
        if (result == currentConstraint) {
            throw new IllegalStateException("Failed to reduce constraint formula");
        }
        if (result != null) {
            if ((result instanceof ConstraintFormula)) {
                if (!this.reduceOneConstraint(context, (ConstraintFormula)result)) {
                    return false;
                }
            }
            else if ((result instanceof ConstraintFormula[])) {
                ConstraintFormula[] resultArray = (ConstraintFormula[])result;
                int i = 0;
                while (i < resultArray.length) {
                    if (!this.reduceOneConstraint(context, resultArray[i])) {
                        return false;
                    }
                    ++i;
                }
            }
            else {
                this.addBound((TypeBound)result, context.environment);
            }
        }
        return true;
    }
    
    public boolean dependsOnResolutionOf(InferenceVariable alpha, InferenceVariable beta) {
        alpha = alpha.prototype();
        beta = beta.prototype();
        if (TypeBinding.equalsEquals(alpha, beta)) {
            return true;
        }
        Iterator<Map.Entry<ParameterizedTypeBinding, ParameterizedTypeBinding>> captureIter = this.captures.entrySet().iterator();
        boolean betaIsInCaptureLhs = false;
        while (captureIter.hasNext()) {
            Map.Entry<ParameterizedTypeBinding, ParameterizedTypeBinding> entry = (Map.Entry)captureIter.next();
            ParameterizedTypeBinding g = (ParameterizedTypeBinding)entry.getKey();
            for (int i = 0; i < g.arguments.length; ++i) {
                if (TypeBinding.equalsEquals(g.arguments[i], alpha)) {
                    ParameterizedTypeBinding captured = (ParameterizedTypeBinding)entry.getValue();
                    if (captured.mentionsAny(new TypeBinding[] { beta }, -1)) {
                        return true;
                    }
                    if (g.mentionsAny(new TypeBinding[] { beta }, i)) {
                        return true;
                    }
                }
                else if (TypeBinding.equalsEquals(g.arguments[i], beta)) {
                    betaIsInCaptureLhs = true;
                }
            }
        }
        if (betaIsInCaptureLhs) {
            BoundSet.ThreeSets sets = (BoundSet.ThreeSets)this.boundsPerVariable.get(beta);
            if (sets != null && sets.hasDependency(alpha)) {
                return true;
            }
        }
        else {
            BoundSet.ThreeSets sets = (BoundSet.ThreeSets)this.boundsPerVariable.get(alpha);
            if (sets != null && sets.hasDependency(beta)) {
                return true;
            }
        }
        return false;
    }
    
    List<Set<InferenceVariable>> computeConnectedComponents(InferenceVariable[] inferenceVariables) {
        Map<InferenceVariable, Set<InferenceVariable>> allEdges = new HashMap();
        for (int i = 0; i < inferenceVariables.length; ++i) {
            InferenceVariable iv1 = inferenceVariables[i];
            HashSet<InferenceVariable> targetSet = new HashSet();
            allEdges.put(iv1, targetSet);
            for (InferenceVariable iv2 : inferenceVariables) {
                if (this.dependsOnResolutionOf(iv1, iv2) || this.dependsOnResolutionOf(iv2, iv1)) {
                    targetSet.add(iv2);
                    ((Set)allEdges.get(iv2)).add(iv1);
                }
            }
        }
        Set<InferenceVariable> visited = new HashSet();
        List<Set<InferenceVariable>> allComponents = new ArrayList();
        for (InferenceVariable inferenceVariable : inferenceVariables) {
            Set<InferenceVariable> component = new HashSet();
            this.addConnected(component, inferenceVariable, allEdges, visited);
            if (!component.isEmpty()) {
                allComponents.add(component);
            }
        }
        return allComponents;
    }
    
    private void addConnected(Set<InferenceVariable> component, InferenceVariable seed, Map<InferenceVariable, Set<InferenceVariable>> allEdges, Set<InferenceVariable> visited) {
        if (visited.add(seed)) {
            component.add(seed);
            for (InferenceVariable next : (Set)allEdges.get(seed)) {
                this.addConnected(component, next, allEdges, visited);
            }
        }
    }
    
    public boolean hasCaptureBound(Set<InferenceVariable> variableSet) {
        for (ParameterizedTypeBinding g : this.captures.keySet()) {
            int i = 0;
            while (i < g.arguments.length) {
                if (variableSet.contains(g.arguments[i])) {
                    return true;
                }
                ++i;
            }
        }
        return false;
    }
    
    public boolean hasOnlyTrivialExceptionBounds(InferenceVariable variable, TypeBinding[] upperBounds) {
        if (upperBounds != null) {
            int i = 0;
            while (i < upperBounds.length) {
                switch (upperBounds[i].id) {
                    case 1:
                    case 21:
                    case 25: {
                        ++i;
                        continue;
                    }
                    default: {
                        return false;
                    }
                }
            }
        }
        return true;
    }
    
    public TypeBinding[] upperBounds(InferenceVariable variable, boolean onlyProper) {
        BoundSet.ThreeSets three = (BoundSet.ThreeSets)this.boundsPerVariable.get(variable.prototype());
        if (three == null || three.subBounds == null) {
            return Binding.NO_TYPES;
        }
        return three.upperBounds(onlyProper, variable);
    }
    
    TypeBinding[] lowerBounds(InferenceVariable variable, boolean onlyProper) {
        BoundSet.ThreeSets three = (BoundSet.ThreeSets)this.boundsPerVariable.get(variable.prototype());
        if (three == null || three.superBounds == null) {
            return Binding.NO_TYPES;
        }
        return three.lowerBounds(onlyProper, variable);
    }
    
    @Override
    public String toString() {
        StringBuffer buf = new StringBuffer("Type Bounds:\n");
        TypeBound[] flattened = this.flatten();
        for (int i = 0; i < flattened.length; ++i) {
            buf.append('\t').append(flattened[i].toString()).append('\n');
        }
        buf.append("Capture Bounds:\n");
        for (Map.Entry<ParameterizedTypeBinding, ParameterizedTypeBinding> capt : this.captures.entrySet()) {
            String lhs = String.valueOf(((TypeBinding)capt.getKey()).shortReadableName());
            String rhs = String.valueOf(((TypeBinding)capt.getValue()).shortReadableName());
            buf.append('\t').append(lhs).append(" = capt(").append(rhs).append(")\n");
        }
        return buf.toString();
    }
    
    public TypeBinding findWrapperTypeBound(InferenceVariable variable) {
        BoundSet.ThreeSets three = (BoundSet.ThreeSets)this.boundsPerVariable.get(variable.prototype());
        if (three == null) {
            return null;
        }
        return three.findSingleWrapperType();
    }
    
    public boolean condition18_5_2_bullet_3_3_1(InferenceVariable alpha, TypeBinding targetType) {
        if (targetType.isBaseType()) {
            return false;
        }
        if (InferenceContext18.parameterizedWithWildcard(targetType) != null) {
            return false;
        }
        BoundSet.ThreeSets ts = (BoundSet.ThreeSets)this.boundsPerVariable.get(alpha.prototype());
        if (ts == null) {
            return false;
        }
        if (ts.sameBounds != null) {
            for (TypeBound bound : ts.sameBounds) {
                if (InferenceContext18.parameterizedWithWildcard(bound.right) != null) {
                    return true;
                }
                continue;
            }
        }
        if (ts.superBounds != null) {
            for (TypeBound bound : ts.superBounds) {
                if (InferenceContext18.parameterizedWithWildcard(bound.right) != null) {
                    return true;
                }
                continue;
            }
        }
        if (ts.superBounds != null) {
            ArrayList<TypeBound> superBounds = new ArrayList(ts.superBounds);
            for (int len = superBounds.size(), i = 0; i < len; ++i) {
                TypeBinding s1 = ((TypeBound)superBounds.get(i)).right;
                int j = i + 1;
                while (j < len) {
                    TypeBinding s2 = ((TypeBound)superBounds.get(j)).right;
                    TypeBinding[] supers = this.superTypesWithCommonGenericType(s1, s2);
                    if (supers != null && supers[0].isProperType(true) && supers[1].isProperType(true) && !TypeBinding.equalsEquals(supers[0], supers[1])) {
                        return true;
                    }
                    ++j;
                }
            }
        }
        return false;
    }
    
    public boolean condition18_5_2_bullet_3_3_2(InferenceVariable alpha, TypeBinding targetType, InferenceContext18 ctx18) {
        if (!targetType.isParameterizedType()) {
            return false;
        }
        TypeBinding g = targetType.original();
        BoundSet.ThreeSets ts = (BoundSet.ThreeSets)this.boundsPerVariable.get(alpha.prototype());
        if (ts == null) {
            return false;
        }
        if (ts.sameBounds != null) {
            for (TypeBound b : ts.sameBounds) {
                if (this.superOnlyRaw(g, b.right, ctx18.environment)) {
                    return true;
                }
                continue;
            }
        }
        if (ts.superBounds != null) {
            for (TypeBound b : ts.superBounds) {
                if (this.superOnlyRaw(g, b.right, ctx18.environment)) {
                    return true;
                }
                continue;
            }
        }
        return false;
    }
    
    private boolean superOnlyRaw(TypeBinding g, TypeBinding s, LookupEnvironment env) {
        if ((s instanceof InferenceVariable)) {
            return false;
        }
        TypeBinding superType = s.findSuperTypeOriginatingFrom(g);
        return superType != null && !superType.isParameterizedType() && s.isCompatibleWith(env.convertToRawType(g, false));
    }
    
    protected TypeBinding[] superTypesWithCommonGenericType(TypeBinding s, TypeBinding t) {
        if (s == null || s.id == 1 || t == null || t.id == 1) {
            return null;
        }
        if (TypeBinding.equalsEquals(s.original(), t.original())) {
            return new TypeBinding[] { s, t };
        }
        TypeBinding tSuper = t.findSuperTypeOriginatingFrom(s);
        if (tSuper != null) {
            return new TypeBinding[] { s, tSuper };
        }
        TypeBinding[] result = this.superTypesWithCommonGenericType(s.superclass(), t);
        if (result != null) {
            return result;
        }
        ReferenceBinding[] superInterfaces = s.superInterfaces();
        if (superInterfaces != null) {
            int i = 0;
            while (i < superInterfaces.length) {
                result = this.superTypesWithCommonGenericType(superInterfaces[i], t);
                if (result != null) {
                    return result;
                }
                ++i;
            }
        }
        return null;
    }
    
    public TypeBinding getEquivalentOuterVariable(InferenceVariable variable, InferenceVariable[] outerVariables) {
        BoundSet.ThreeSets three = (BoundSet.ThreeSets)this.boundsPerVariable.get(variable);
        if (three != null) {
            for (TypeBound bound : three.sameBounds) {
                int length = outerVariables.length;
                int i = 0;
                while (i < length) {
                    InferenceVariable iv = outerVariables[i];
                    if (TypeBinding.equalsEquals(bound.right, iv)) {
                        return iv;
                    }
                    ++i;
                }
            }
        }
        for (InferenceVariable iv2 : outerVariables) {
            three = (BoundSet.ThreeSets)this.boundsPerVariable.get(outerVariables);
            if (three != null) {
                for (TypeBound bound2 : three.sameBounds) {
                    if (TypeBinding.equalsEquals(bound2.right, variable)) {
                        return iv2;
                    }
                    continue;
                }
            }
        }
        return null;
    }
    
    private class ThreeSets
    {
        Set<TypeBound> superBounds;
        Set<TypeBound> sameBounds;
        Set<TypeBound> subBounds;
        TypeBinding instantiation;
        Map<InferenceVariable, TypeBound> inverseBounds;
        
        public ThreeSets() {
        }
        
        public ParameterizedTypeBinding mergeTypeParameters(ParameterizedTypeBinding current, ParameterizedTypeBinding newB) {
            TypeBinding[] curTypeArgs = current.typeArguments();
            TypeBinding[] newTypeArgs = newB.typeArguments();
            TypeBinding[] merged = new TypeBinding[curTypeArgs.length];
            System.arraycopy(curTypeArgs, 0, merged, 0, curTypeArgs.length);
            boolean wasMerged = false;
            for (int i = 0; i < curTypeArgs.length; ++i) {
                if (!TypeBinding.equalsEquals(curTypeArgs[i], newTypeArgs[i])) {
                    if (curTypeArgs[i].isCapture() || newTypeArgs[i].isCapture()) {
                        return null;
                    }
                    if ((curTypeArgs[i] instanceof InferenceVariable)) {
                        if (!(newTypeArgs[i] instanceof InferenceVariable)) {
                            BoundSet.ThreeSets three = (BoundSet.ThreeSets)BoundSet.this.boundsPerVariable.get(curTypeArgs[i]);
                            if (three != null && three.sameBounds != null && three.sameBounds.contains(new TypeBound((InferenceVariable)curTypeArgs[i], newTypeArgs[i], 4))) {
                                merged[i] = newTypeArgs[i];
                                wasMerged = true;
                            }
                        }
                        else if (!curTypeArgs[i].equals(newTypeArgs[i])) {
                            return null;
                        }
                    }
                    else if (!(newTypeArgs[i] instanceof InferenceVariable)) {
                        if (!TypeBinding.equalsEquals(curTypeArgs[i], newTypeArgs[i])) {
                            return null;
                        }
                    }
                    else {
                        BoundSet.ThreeSets three = (BoundSet.ThreeSets)BoundSet.this.boundsPerVariable.get(newTypeArgs[i]);
                        if (three == null || three.sameBounds == null || !three.sameBounds.contains(new TypeBound((InferenceVariable)newTypeArgs[i], curTypeArgs[i], 4))) {
                            return null;
                        }
                    }
                }
            }
            if (wasMerged) {
                ParameterizedTypeBinding clone = (ParameterizedTypeBinding)current.clone(current.enclosingType());
                clone.arguments = merged;
                return clone;
            }
            return null;
        }
        
        public boolean addBound(TypeBound bound) {
            Iterator<TypeBound> it = null;
            switch (bound.relation) {
                case 3: {
                    if (this.superBounds == null) {
                        this.superBounds = new HashSet();
                    }
                    if (CompilerOptions.useunspecdtypeinferenceperformanceoptimization && !bound.right.isProperType(true)) {
                        it = this.superBounds.iterator();
                        while (it.hasNext()) {
                            TypeBound b = (TypeBound)it.next();
                            if (bound.right.isParameterizedType() && b.right.isParameterizedType() && b.right.original() == bound.right.original()) {
                                TypeBinding clone = this.mergeTypeParameters((ParameterizedTypeBinding)b.right, (ParameterizedTypeBinding)bound.right);
                                if (clone != null) {
                                    b.right = clone;
                                    return false;
                                }
                                continue;
                            }
                            else {
                                continue;
                            }
                        }
                    }
                    return this.superBounds.add(bound);
                }
                case 4: {
                    if (this.sameBounds == null) {
                        this.sameBounds = new HashSet();
                    }
                    return this.sameBounds.add(bound);
                }
                case 2: {
                    if (this.subBounds == null) {
                        this.subBounds = new HashSet();
                    }
                    if (CompilerOptions.useunspecdtypeinferenceperformanceoptimization && !bound.right.isProperType(true)) {
                        it = this.subBounds.iterator();
                        while (it.hasNext()) {
                            TypeBound b = (TypeBound)it.next();
                            if (bound.right.isParameterizedType() && b.right.isParameterizedType() && b.right.original() == bound.right.original()) {
                                TypeBinding clone = this.mergeTypeParameters((ParameterizedTypeBinding)b.right, (ParameterizedTypeBinding)bound.right);
                                if (clone != null) {
                                    b.right = clone;
                                    return false;
                                }
                                continue;
                            }
                            else {
                                continue;
                            }
                        }
                    }
                    return this.subBounds.add(bound);
                }
                default: {
                    throw new IllegalArgumentException("Unexpected bound relation in : " + bound);
                }
            }
        }
        
        public TypeBinding[] lowerBounds(boolean onlyProper, InferenceVariable variable) {
            TypeBinding[] boundTypes = new TypeBinding[this.superBounds.size()];
            Iterator<TypeBound> it = this.superBounds.iterator();
            long nullHints = variable.nullHints;
            int i = 0;
            while (it.hasNext()) {
                TypeBound current = (TypeBound)it.next();
                TypeBinding boundType = current.right;
                if (!onlyProper || boundType.isProperType(true)) {
                    boundTypes[i++] = boundType;
                    nullHints |= current.nullHints;
                }
                else {
                    continue;
                }
            }
            if (i == 0) {
                return Binding.NO_TYPES;
            }
            if (i < boundTypes.length) {
                System.arraycopy(boundTypes, 0, boundTypes = new TypeBinding[i], 0, i);
            }
            this.useNullHints(nullHints, boundTypes, variable.environment);
            InferenceContext18.sortTypes(boundTypes);
            return boundTypes;
        }
        
        public TypeBinding[] upperBounds(boolean onlyProper, InferenceVariable variable) {
            ReferenceBinding[] rights = new ReferenceBinding[this.subBounds.size()];
            TypeBinding simpleUpper = null;
            Iterator<TypeBound> it = this.subBounds.iterator();
            long nullHints = variable.nullHints;
            int i = 0;
            while (it.hasNext()) {
                TypeBinding right = ((TypeBound)it.next()).right;
                if (!onlyProper || right.isProperType(true)) {
                    if ((right instanceof ReferenceBinding)) {
                        rights[i++] = (ReferenceBinding)right;
                        nullHints |= (right.tagBits & 0x180000000000000L);
                    }
                    else {
                        if (simpleUpper != null) {
                            return Binding.NO_TYPES;
                        }
                        simpleUpper = right;
                    }
                }
                else {
                    continue;
                }
            }
            if (i == 0) {
                return (simpleUpper != null) ? new TypeBinding[] { simpleUpper } : Binding.NO_TYPES;
            }
            if (i == 1 && simpleUpper != null) {
                return new TypeBinding[] { simpleUpper };
            }
            if (i < rights.length) {
                System.arraycopy(rights, 0, rights = new ReferenceBinding[i], 0, i);
            }
            this.useNullHints(nullHints, rights, variable.environment);
            InferenceContext18.sortTypes(rights);
            return rights;
        }
        
        public boolean hasDependency(InferenceVariable beta) {
            return (this.superBounds != null && this.hasDependency(this.superBounds, beta)) || (this.sameBounds != null && this.hasDependency(this.sameBounds, beta)) || (this.subBounds != null && this.hasDependency(this.subBounds, beta)) || (this.inverseBounds != null && this.inverseBounds.containsKey(beta));
        }
        
        private boolean hasDependency(Set<TypeBound> someBounds, InferenceVariable var) {
            for (TypeBound bound : someBounds) {
                if (TypeBinding.equalsEquals(bound.right, var) || bound.right.mentionsAny(new TypeBinding[] { var }, -1)) {
                    return true;
                }
                continue;
            }
            return false;
        }
        
        public int size() {
            int size = 0;
            if (this.superBounds != null) {
                size += this.superBounds.size();
            }
            if (this.sameBounds != null) {
                size += this.sameBounds.size();
            }
            if (this.subBounds != null) {
                size += this.subBounds.size();
            }
            return size;
        }
        
        public int flattenInto(TypeBound[] collected, int idx) {
            if (this.superBounds != null) {
                int len = this.superBounds.size();
                System.arraycopy(this.superBounds.toArray(), 0, collected, idx, len);
                idx += len;
            }
            if (this.sameBounds != null) {
                int len = this.sameBounds.size();
                System.arraycopy(this.sameBounds.toArray(), 0, collected, idx, len);
                idx += len;
            }
            if (this.subBounds != null) {
                int len = this.subBounds.size();
                System.arraycopy(this.subBounds.toArray(), 0, collected, idx, len);
                idx += len;
            }
            return idx;
        }
        
        public BoundSet.ThreeSets copy() {
            BoundSet.ThreeSets copy = new BoundSet.ThreeSets();
            if (this.superBounds != null) {
                copy.superBounds = new HashSet(this.superBounds);
            }
            if (this.sameBounds != null) {
                copy.sameBounds = new HashSet(this.sameBounds);
            }
            if (this.subBounds != null) {
                copy.subBounds = new HashSet(this.subBounds);
            }
            copy.instantiation = this.instantiation;
            return copy;
        }
        
        public TypeBinding findSingleWrapperType() {
            if (this.instantiation != null && this.instantiation.isProperType(true)) {
                switch (this.instantiation.id) {
                    case 26:
                    case 27:
                    case 28:
                    case 29:
                    case 30:
                    case 31:
                    case 32:
                    case 33: {
                        return this.instantiation;
                    }
                    default: {
                        break;
                    }
                }
            }
            if (this.subBounds != null) {
                Iterator<TypeBound> it = this.subBounds.iterator();
                while (it.hasNext()) {
                    TypeBinding boundType = ((TypeBound)it.next()).right;
                    if (boundType.isProperType(true)) {
                        switch (boundType.id) {
                            case 26:
                            case 27:
                            case 28:
                            case 29:
                            case 30:
                            case 31:
                            case 32:
                            case 33: {
                                return boundType;
                            }
                            default: {
                                continue;
                            }
                        }
                    }
                    else {
                        continue;
                    }
                }
            }
            if (this.superBounds != null) {
                Iterator<TypeBound> it = this.superBounds.iterator();
                while (it.hasNext()) {
                    TypeBinding boundType = ((TypeBound)it.next()).right;
                    if (boundType.isProperType(true)) {
                        switch (boundType.id) {
                            case 26:
                            case 27:
                            case 28:
                            case 29:
                            case 30:
                            case 31:
                            case 32:
                            case 33: {
                                return boundType;
                            }
                            default: {
                                continue;
                            }
                        }
                    }
                    else {
                        continue;
                    }
                }
            }
            return null;
        }
        
        private void useNullHints(long nullHints, TypeBinding[] boundTypes, LookupEnvironment environment) {
            if (nullHints == 108086391056891904L) {
                for (int j = 0; j < boundTypes.length; ++j) {
                    boundTypes[j] = boundTypes[j].withoutToplevelNullAnnotation();
                }
            }
            else {
                AnnotationBinding[] annot = environment.nullAnnotationsFromTagBits(nullHints);
                if (annot != null) {
                    for (int i = 0; i < boundTypes.length; ++i) {
                        boundTypes[i] = environment.createAnnotatedType(boundTypes[i], annot);
                    }
                }
            }
        }
        
        TypeBinding combineAndUseNullHints(TypeBinding type, long nullHints, LookupEnvironment environment) {
            if (this.sameBounds != null) {
                Iterator<TypeBound> it = this.sameBounds.iterator();
                while (it.hasNext()) {
                    nullHints |= ((TypeBound)it.next()).nullHints;
                }
            }
            if (this.superBounds != null) {
                Iterator<TypeBound> it = this.superBounds.iterator();
                while (it.hasNext()) {
                    nullHints |= ((TypeBound)it.next()).nullHints;
                }
            }
            if (this.subBounds != null) {
                Iterator<TypeBound> it = this.subBounds.iterator();
                while (it.hasNext()) {
                    nullHints |= ((TypeBound)it.next()).nullHints;
                }
            }
            if (nullHints == 108086391056891904L) {
                return type.withoutToplevelNullAnnotation();
            }
            AnnotationBinding[] annot = environment.nullAnnotationsFromTagBits(nullHints);
            if (annot != null) {
                return environment.createAnnotatedType(type, annot);
            }
            return type;
        }
        
        public void setInstantiation(TypeBinding type, InferenceVariable variable, LookupEnvironment environment) {
            if (environment.globalOptions.isAnnotationBasedNullAnalysisEnabled) {
                long variableBits = variable.tagBits & 0x180000000000000L;
                long allBits = type.tagBits | variableBits;
                if (this.instantiation != null) {
                    allBits |= this.instantiation.tagBits;
                }
                allBits &= 0x180000000000000L;
                if (allBits == 108086391056891904L) {
                    allBits = variableBits;
                }
                if (allBits != (type.tagBits & 0x180000000000000L)) {
                    AnnotationBinding[] annot = environment.nullAnnotationsFromTagBits(allBits);
                    if (annot != null) {
                        type = environment.createAnnotatedType(type.withoutToplevelNullAnnotation(), annot);
                    }
                    else if (type.hasNullTypeAnnotations()) {
                        type = type.withoutToplevelNullAnnotation();
                    }
                }
            }
            this.instantiation = type;
        }
    }
}
